const uuid = require('uuid')
const { pick, get, set, has, isEmpty, last, findIndex } = require('lodash')
const config = require('config')
const { v4 } = require('uuid')

const {
  services,
  Fragment,
  Collection,
  authsome: authsomeHelper,
} = require('pubsweet-component-helper-service')

const { features = {}, recommendations } = config

const sendMTSPackage = async (collection, fragment, isEQA = false) => {
  const s3Config = get(config, 'pubsweet-component-aws-s3', {})
  const mtsConfig = get(config, 'mts-service', {})
  const MTSService = require('pubsweet-component-mts-package')

  const { journal, xmlParser, ftp } = mtsConfig
  const MTS = new MTSService(journal, xmlParser, s3Config, ftp)
  const packageFragment = {
    ...fragment,
    metadata: {
      ...fragment.metadata,
      customId: collection.customId,
    },
  }

  await MTS.sendPackage({ fragment: packageFragment, isEQA })
}

const Notification = require('../../notifications/notification')

module.exports = models => async (req, res) => {
  const { recommendation, comments, recommendationType } = req.body
  if (!services.checkForUndefinedParams(recommendationType))
    return res.status(400).json({ error: 'Recommendation type is required.' })

  const reqUser = await models.User.find(req.user)
  const isEditorInChief = reqUser.editorInChief || reqUser.admin

  const { collectionId, fragmentId } = req.params

  let collection, fragment

  try {
    collection = await models.Collection.find(collectionId)
    if (!collection.fragments.includes(fragmentId))
      return res.status(400).json({
        error: `Collection and fragment do not match.`,
      })
    fragment = await models.Fragment.find(fragmentId)
  } catch (e) {
    const notFoundError = await services.handleNotFoundError(e, 'Item')
    return res.status(notFoundError.status).json({
      error: notFoundError.message,
    })
  }

  const authsome = authsomeHelper.getAuthsome(models)
  const target = {
    fragment,
    path: req.route.path,
  }
  const canPost = await authsome.can(req.user, 'POST', target)
  if (!canPost)
    return res.status(403).json({
      error: 'Unauthorized.',
    })

  if (
    recommendationType === recommendations.type.editor &&
    last(collection.fragments) !== fragmentId
  ) {
    return res
      .status(400)
      .json({ error: 'Cannot make a recommendation on an older version.' })
  }

  if (
    recommendationType === recommendations.type.review &&
    last(collection.fragments) !== fragmentId
  ) {
    return res
      .status(400)
      .json({ error: 'Cannot write a review on an older version.' })
  }

  if (
    recommendation === recommendations.publish &&
    recommendationType === recommendations.type.editor &&
    collection.handlingEditor &&
    collection.handlingEditor.id === req.user
  ) {
    const fragmentHelper = new Fragment({ fragment })
    if (!fragmentHelper.hasReviewReport()) {
      return res
        .status(400)
        .json({ error: 'Cannot publish without at least one reviewer report.' })
    }
  }

  fragment.recommendations = fragment.recommendations || []
  const newRecommendation = {
    id: uuid.v4(),
    userId: reqUser.id,
    createdOn: Date.now(),
    updatedOn: Date.now(),
    recommendationType,
  }

  newRecommendation.recommendation = recommendation || undefined
  newRecommendation.comments = comments || undefined

  const collectionHelper = new Collection({ collection })
  if (recommendationType === 'editorRecommendation') {
    collectionHelper.updateStatusOnRecommendation({
      isEditorInChief,
      recommendation,
    })

    if (!isEditorInChief && ['minor', 'major'].includes(recommendation)) {
      fragment.revision = pick(fragment, ['authors', 'files', 'metadata'])
    }

    const technicalChecks = get(collection, 'technicalChecks', {})
    const hasEQA = has(technicalChecks, 'eqa')
    // the manuscript has not yet passed through the EQA process so we need to upload it to the FTP server
    if (isEditorInChief && recommendation === 'publish' && !hasEQA) {
      if (features.mts) {
        await sendMTSPackage(collection, fragment, true)
      }

      collection.status = 'inQA'
      set(collection, 'technicalChecks.token', v4())
      set(collection, 'technicalChecks.eqa', false)
      await collection.save()
    }

    /* if the EiC returns the manuscript to the HE after the EQA has been performed
       then remove all properties from the technicalChecks property so that the manuscript
       can go through the EQA process again
    */
    if (
      isEditorInChief &&
      recommendation === 'return-to-handling-editor' &&
      hasEQA
    ) {
      collection.technicalChecks = {}
      await collection.save()
    }

    const notification = new Notification({
      fragment,
      collection,
      newRecommendation,
      UserModel: models.User,
      baseUrl: services.getBaseUrl(req),
    })

    const hasPeerReview = !isEmpty(collection.handlingEditor)

    if (isEditorInChief) {
      if (recommendation === 'publish' && collection.status === 'inQA') {
        notification.notifyEAWhenEiCRequestsEQAApproval()
      }

      if (recommendation === 'publish' && collection.status === 'accepted') {
        notification.notifyEAWhenEiCMakesFinalDecision()
      }

      if (hasPeerReview && (recommendation !== 'publish' || hasEQA)) {
        if (recommendation === 'return-to-handling-editor') {
          notification.notifyHEWhenEiCReturnsToHE()
        } else {
          notification.notifyHEWhenEiCMakesDecision()
          notification.notifyReviewersWhenEiCMakesDecision()
        }
      }

      if (
        recommendation !== 'return-to-handling-editor' &&
        (recommendation !== 'publish' || hasEQA)
      ) {
        notification.notifyAuthorsWhenEiCMakesDecision()
      }
    } else {
      if (collection.status === 'revisionRequested') {
        notification.notifySAWhenHERequestsRevision()
      }

      if (hasPeerReview) {
        notification.notifyReviewersWhenHEMakesRecommendation()
        notification.notifyEiCWhenHEMakesRecommendation()
      }
    }
  }

  const invitationIndex = findIndex(fragment.invitations, [
    'userId',
    reqUser.id,
  ])
  if (invitationIndex >= 0) {
    const FragmentModel = models.Fragment
    const reviewerNumber = await collectionHelper.getReviewerNumber({
      reqUser,
      FragmentModel,
    })
    fragment.invitations[invitationIndex].reviewerNumber = reviewerNumber
  }
  fragment.recommendations.push(newRecommendation)
  fragment.save()

  return res.status(200).json(newRecommendation)
}
